/***************************************************************************
 *   Copyright (c) 2004 Jrgen Riegel <juergen.riegel@web.de>              *
 *                                                                         *
 *   This file is part of the FreeCAD CAx development system.              *
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Library General Public           *
 *   License as published by the Free Software Foundation; either          *
 *   version 2 of the License, or (at your option) any later version.      *
 *                                                                         *
 *   This library  is distributed in the hope that it will be useful,      *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Library General Public License for more details.                  *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this library; see the file COPYING.LIB. If not,    *
 *   write to the Free Software Foundation, Inc., 59 Temple Place,         *
 *   Suite 330, Boston, MA  02111-1307, USA                                *
 *                                                                         *
 ***************************************************************************/


#include "PreCompiled.h"

#ifndef _PreComp_
# include <QApplication>
# include <QFileInfo>
# include <QKeyEvent>
# include <QEvent>
# include <QDropEvent>
# include <QDragEnterEvent>
# include <Inventor/actions/SoWriteAction.h>
# include <Inventor/actions/SoGetPrimitiveCountAction.h>
# include <Inventor/nodes/SoMaterial.h>
# include <Inventor/nodes/SoOrthographicCamera.h>
# include <Inventor/nodes/SoPerspectiveCamera.h>
# include <Inventor/nodes/SoSeparator.h>
# include <Inventor/nodes/SoShapeHints.h>
# include <Inventor/events/SoEvent.h>
# include <Inventor/fields/SoSFString.h>
# include <Inventor/fields/SoSFColor.h>
#endif

#include <Base/Exception.h>

#include "View3DInventor.h"
#include "View3DInventorViewer.h"
#include "Document.h"
#include "Application.h"
#include "MainWindow.h"
#include "MenuManager.h"
#include <App/Feature.h>
#include <Base/Console.h>

// build in Inventor
#include "Inventor/Qt/viewers/SoQtExaminerViewer.h"
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodes/SoOrthographicCamera.h>

#include "View3DInventorExamples.h"
#include "SoFCSelectionAction.h"
#include "View3DPy.h"

#include <locale>

using namespace Gui;

TYPESYSTEM_SOURCE_ABSTRACT(Gui::View3DInventor,Gui::BaseView);

View3DInventor::View3DInventor(Gui::Document* pcDocument, QWidget* parent, Qt::WFlags wflags)
    : MDIView(pcDocument, parent, wflags), _viewerPy(0)
{
    // important for highlighting 
    setMouseTracking(true);
    // accept drops on the window, get handled in dropEvent, dragEnterEvent   
    setAcceptDrops(true);
  
    // attache Parameter Observer
    hGrp = App::GetApplication().GetParameterGroupByPath("User parameter:BaseApp/Preferences/View");
    hGrp->Attach(this);

    // create the inventor widget and set the defaults
    _viewer = new View3DInventorViewer(this);
    setViewerDefaults();
    // check whether a perspective or orthogrphic camera should be set
    if ( hGrp->GetBool("Orthographic", false) )
        _viewer->setCameraType(SoOrthographicCamera::getClassTypeId());
    else
        _viewer->setCameraType(SoPerspectiveCamera::getClassTypeId());
 
    // check whether the simple or the Full Mouse model is used
    _viewer->setMouseModel(hGrp->GetInt("MouseModel",1));

    stopSpinTimer = new QTimer(this);
    connect(stopSpinTimer, SIGNAL(timeout()), this, SLOT(stopAnimating()));
}

View3DInventor::~View3DInventor()
{
    hGrp->Detach(this);

    //If we destroy this viewer by calling 'delete' directly the focus proxy widget which is defined 
    //by a widget in SoQtViewer isn't resetted. This widget becomes to a dangling pointer and makes
    //the application crash. (Probably it's better to destroy this viewer by calling close().)
    //See also Gui::Document::~Document().
    QWidget* foc = qApp->focusWidget();
    if (foc) {
        QWidget* par = foc->parentWidget();
        while (par) {
            if (par == this) {
                foc->setFocusProxy(0);
                foc->clearFocus();
                break;
            }
            par = par->parentWidget();
        }
    }

    if (_viewerPy) {
        _viewerPy->setInvalid();
        Py_DECREF(_viewerPy);
    }

    delete _viewer;
}

PyObject *View3DInventor::getPyObject(void)
{
    if (!_viewerPy)
        _viewerPy = new View3DPy(this);

    Py_INCREF(_viewerPy);
    return _viewerPy;
}

void View3DInventor::setViewerDefaults(void)
{
    _viewer->setStereoOffset(hGrp->GetFloat("EyeDistance"      ,65.0));
    _viewer->bDrawAxisCross = hGrp->GetBool("CornerCoordSystem",true);
    _viewer->bAllowSpining =  hGrp->GetBool("UseAutoRotation"  ,true);
    _viewer->setGradientBackgroud( (hGrp->GetBool("Gradient",true)) );
    unsigned long col = hGrp->GetUnsigned("BackgroundColor",0);
    float r,g,b;
    r = ((col >> 24) & 0xff) / 255.0; g = ((col >> 16) & 0xff) / 255.0; b = ((col >> 8) & 0xff) / 255.0;
    _viewer->setBackgroundColor(SbColor(r, g, b));

    unsigned long col2 = hGrp->GetUnsigned("BackgroundColor2",2139082239); // default color (lila)
    unsigned long col3 = hGrp->GetUnsigned("BackgroundColor3",ULONG_MAX); // default color (white)
    float r2,g2,b2,r3,g3,b3;
    r2 = ((col2 >> 24) & 0xff) / 255.0; g2 = ((col2 >> 16) & 0xff) / 255.0; b2 = ((col2 >> 8) & 0xff) / 255.0;
    r3 = ((col3 >> 24) & 0xff) / 255.0; g3 = ((col3 >> 16) & 0xff) / 255.0; b3 = ((col3 >> 8) & 0xff) / 255.0;
    _viewer->setGradientBackgroudColor( SbColor(r2, g2, b2), SbColor(r3, g3, b3) );

    if (hGrp->GetBool("UseAntialiasing"  ,false))
        _viewer->getGLRenderAction()->setSmoothing(true);
    else
        _viewer->getGLRenderAction()->setSmoothing(false);
    _viewer->setEnabledFPSCounter(hGrp->GetBool("ShowFPS",false));
}

void View3DInventor::OnChange(ParameterGrp::SubjectType &rCaller,ParameterGrp::MessageType Reason)
{
    if ( strcmp(Reason,"DisablePreselection") == 0 ) {
        const ParameterGrp& rclGrp = ((ParameterGrp&)rCaller);
        SoFCEnableHighlightAction cAct( !rclGrp.GetBool("DisablePreselection", false) );
        cAct.apply(_viewer->getSceneGraph());
    } else if ( strcmp(Reason,"DisableSelection") == 0) {
        const ParameterGrp& rclGrp = ((ParameterGrp&)rCaller);
        SoFCEnableSelectionAction cAct( !rclGrp.GetBool("DisableSelection", false) );
        cAct.apply(_viewer->getSceneGraph());
    } else if ( strcmp(Reason,"HighlightColor") == 0) {
        float transparency;
        SbColor highlightColor(0.1f, 0.1f, 0.8f);
        unsigned long highlight = (unsigned long)(highlightColor.getPackedValue());
        int a = (highlight)&0xff;
        highlight = hGrp->GetUnsigned("HighlightColor", highlight);
        highlight += a;
        highlightColor.setPackedValue((uint32_t)highlight, transparency);
        SoSFColor col; col.setValue(highlightColor);
        SoFCHighlightColorAction cAct( col );
        cAct.apply(_viewer->getSceneGraph());
    } else if ( strcmp(Reason,"SelectionColor") == 0) {
        float transparency;
        SbColor selectionColor(0.1f, 0.5f, 0.1f);
        unsigned long selection = (unsigned long)(selectionColor.getPackedValue());
        int a = (selection)&0xff;
        selection = hGrp->GetUnsigned("SelectionColor", selection);
        selection += a;
        selectionColor.setPackedValue((uint32_t)selection, transparency);
        SoSFColor col; col.setValue(selectionColor);
        SoFCSelectionColorAction cAct( col );
        cAct.apply(_viewer->getSceneGraph());
    } else if ( strcmp(Reason,"MouseModel") == 0 ) {
        // check whether the simple or the Full Mouse model is used
        const ParameterGrp& rclGrp = ((ParameterGrp&)rCaller);
        int model = rclGrp.GetInt("MouseModel",1);
        _viewer->setMouseModel(model);
    } else {
        setViewerDefaults();
    }
}

void View3DInventor::onRename(Gui::Document *pDoc)
{
    SoSFString name;
    name.setValue(pDoc->getDocument()->getName());
    SoFCDocumentAction cAct(name);
    cAct.apply(_viewer->getSceneGraph());
}

void View3DInventor::onUpdate(void)
{
#ifdef FC_LOGUPDATECHAIN
    Base::Console().Log("Acti: Gui::View3DInventor::onUpdate()");
#endif
    update();  
    _viewer->render();
}

void View3DInventor::viewAll()
{
    _viewer->viewAll();
}

const char *View3DInventor::getName(void) const
{
    return "View3DInventor";
}

// buffer acrobatics for inventor ****************************************************
static char * buffer;
static size_t buffer_size = 0;
static std::string cReturnString;

static void *
buffer_realloc(void * bufptr, size_t size)
{
    buffer = (char *)realloc(bufptr, size);
    buffer_size = size;
    return buffer;
}

const std::string &View3DInventor::writeNodesToString(SoNode * root)
{
    SoOutput out;
    buffer = (char *)malloc(1024);
    buffer_size = 1024;
    out.setBuffer(buffer, buffer_size, buffer_realloc);

    SoWriteAction wa(&out);
    wa.apply(root);

    cReturnString =buffer;
    free(buffer);
    return cReturnString;
}

// **********************************************************************************

bool View3DInventor::onMsg(const char* pMsg, const char** ppReturn)
{
  if(strcmp("ViewFit",pMsg) == 0 ){
    _viewer->viewAll();
    return true;
  }else if(strcmp("ViewSelection",pMsg) == 0) {
    _viewer->viewSelection();
    return true;
// comment out on older Inventor
#if (SOQT_MAJOR_VERSION >= 1 && SOQT_MINOR_VERSION >= 2)
  }else if(strcmp("SetStereoRedGreen",pMsg) == 0 ){
    _viewer->setStereoType(SoQtViewer::STEREO_ANAGLYPH);
    return true;
  }else if(strcmp("SetStereoQuadBuff",pMsg) == 0 ){
    _viewer->setStereoType(SoQtViewer::STEREO_QUADBUFFER );
    return true;
  }else if(strcmp("SetStereoInterleavedRows",pMsg) == 0 ){
    _viewer->setStereoType(SoQtViewer::STEREO_INTERLEAVED_ROWS );
    return true;
  }else if(strcmp("SetStereoInterleavedColumns",pMsg) == 0 ){
    _viewer->setStereoType(SoQtViewer::STEREO_INTERLEAVED_COLUMNS  );
    return true;
  }else if(strcmp("SetStereoOff",pMsg) == 0 ){
    _viewer->setStereoType(SoQtViewer::STEREO_NONE );
    return true;
#else
  }else if(strcmp("SetStereoRedGreen",pMsg) == 0 ){
    Base::Console().Warning("Use SoQt 1.2.x or later!\n");
    return true;
  }else if(strcmp("SetStereoQuadBuff",pMsg) == 0 ){
    Base::Console().Warning("Use SoQt 1.2.x or later!\n");
    return true;
  }else if(strcmp("SetStereoInterleavedRows",pMsg) == 0 ){
    Base::Console().Warning("Use SoQt 1.2.x or later!\n");
    return true;
  }else if(strcmp("SetStereoInterleavedColumns",pMsg) == 0 ){
    Base::Console().Warning("Use SoQt 1.2.x or later!\n");
    return true;
  }else if(strcmp("SetStereoOff",pMsg) == 0 ){
    Base::Console().Warning("Use SoQt 1.2.x or later!\n");
    return true;
#endif
  }else if(strcmp("Example1",pMsg) == 0 ){
    SoSeparator * root = new SoSeparator;
    Texture3D(root);
    _viewer->setSceneGraph(root);
    return true;
  }else if(strcmp("Example2",pMsg) == 0 ){
    SoSeparator * root = new SoSeparator;
    LightManip(root);
    _viewer->setSceneGraph(root);
    return true;
  }else if(strcmp("Example3",pMsg) == 0 ){
    SoSeparator * root = new SoSeparator;
    AnimationTexture(root);
    _viewer->setSceneGraph(root);
    return true;
  }else if(strcmp("GetCamera",pMsg) == 0 ){
    SoCamera * Cam = _viewer->getCamera();
    *ppReturn = writeNodesToString(Cam).c_str();
    return true;
  }else if(strncmp("SetCamera",pMsg,9) == 0 ){
    return setCamera(pMsg+10);
  }else if(strncmp("Dump",pMsg,4) == 0 ){
    dump(pMsg+5);
    return true;
  }else if(strcmp("ViewBottom",pMsg) == 0 ){
    SoCamera* cam = _viewer->getCamera();
    cam->orientation.setValue(-1, 0, 0, 0);
    _viewer->viewAll();
    return true;
  }else if(strcmp("ViewFront",pMsg) == 0 ){
    float root = (float)(sqrt(2.0)/2.0);
    SoCamera* cam = _viewer->getCamera();
    cam->orientation.setValue(-root, 0, 0, -root);
    _viewer->viewAll();
    return true;
  }else if(strcmp("ViewLeft",pMsg) == 0 ){
    SoCamera* cam = _viewer->getCamera();
    cam->orientation.setValue(0.5, 0.5, 0.5, 0.5);
    _viewer->viewAll();
    return true;
  }else if(strcmp("ViewRear",pMsg) == 0 ){
    float root = (float)(sqrt(2.0)/2.0);
    SoCamera* cam = _viewer->getCamera();
    cam->orientation.setValue(0, root, root, 0);
    _viewer->viewAll();
    return true;
  }else if(strcmp("ViewRight",pMsg) == 0 ){
    SoCamera* cam = _viewer->getCamera();
    cam->orientation.setValue(-0.5, 0.5, 0.5, -0.5);
    _viewer->viewAll();
    return true;
  }else if(strcmp("ViewTop",pMsg) == 0 ){
    SoCamera* cam = _viewer->getCamera();
    cam->orientation.setValue(0, 0, 0, 1);
    _viewer->viewAll();
    return true;
  }else if(strcmp("ViewAxo",pMsg) == 0 ){
    float root = (float)(sqrt(3.0)/4.0);
    SoCamera* cam = _viewer->getCamera();
    cam->orientation.setValue(-0.333333f, -0.166666f, -0.333333f, -root);
    _viewer->viewAll();
    return true;
  }else if(strcmp("OrthographicCamera",pMsg) == 0 ){
    _viewer->setCameraType(SoOrthographicCamera::getClassTypeId());
    return true;
  }else if(strcmp("PerspectiveCamera",pMsg) == 0 ){
    _viewer->setCameraType(SoPerspectiveCamera::getClassTypeId());
    return true;
  }else  if(strcmp("Undo",pMsg) == 0 ){
    getGuiDocument()->undo(1);
    return true;
  }else  if(strcmp("Redo",pMsg) == 0 ){
    getGuiDocument()->redo(1);
    return true;
  }else if (strcmp("Save",pMsg) == 0){
    getGuiDocument()->save();
    return true;
  }else if (strcmp("SaveAs",pMsg) == 0){
    getGuiDocument()->saveAs();
    return true;
  }else 

  return false;
}


bool View3DInventor::onHasMsg(const char* pMsg) const
{
  if(strcmp("Save",pMsg) == 0 ){
    return true;
  }else if(strcmp("SaveAs",pMsg) == 0 ){
    return true;
  }else if(strcmp("Undo",pMsg) == 0 ){
    return getAppDocument()->getAvailableUndos() > 0;
  }else if(strcmp("Redo",pMsg) == 0 ){
    return getAppDocument()->getAvailableRedos() > 0; 
  }else if(strcmp("SetStereoRedGreen",pMsg) == 0 ){
    return true;
  }else if(strcmp("SetStereoQuadBuff",pMsg) == 0 ){
    return true;
  }else if(strcmp("SetStereoInterleavedRows",pMsg) == 0 ){
    return true;
  }else if(strcmp("SetStereoInterleavedColumns",pMsg) == 0 ){
    return true;
  }else if(strcmp("SetStereoOff",pMsg) == 0 ){
    return true;
  }else if(strcmp("Example1",pMsg) == 0 ){
    return true;
  }else if(strcmp("Example2",pMsg) == 0 ){
    return true;
  }else if(strcmp("Example3",pMsg) == 0 ){
    return true;
  }else if(strcmp("ViewFit",pMsg) == 0 ){
    return true;
  }else if(strcmp("ViewSelection",pMsg) == 0 ){
    return true;
  }else if(strcmp("ViewBottom",pMsg) == 0 ){
    return true;
  }else if(strcmp("ViewFront",pMsg) == 0 ){
    return true;
  }else if(strcmp("ViewLeft",pMsg) == 0 ){
    return true;
  }else if(strcmp("ViewRear",pMsg) == 0 ){
    return true;
  }else if(strcmp("ViewRight",pMsg) == 0 ){
    return true;
  }else if(strcmp("ViewTop",pMsg) == 0 ){
    return true;
  }else if(strcmp("ViewAxo",pMsg) == 0 ){
    return true;
  }else if(strcmp("GetCamera",pMsg) == 0 ){
    return true;
  }else if(strncmp("SetCamera",pMsg,9) == 0 ){
    return true;
  }else if(strncmp("Dump",pMsg,4) == 0 ){
    return true;
  }
  return false;
}

bool View3DInventor::setCamera(const char* pCamera)
{
    SoCamera * CamViewer = _viewer->getCamera();
    if (!CamViewer) {
        throw Base::Exception("No camera set so far...");
    }

    SoInput in;
    in.setBuffer((void*)pCamera,std::strlen(pCamera));

    SoNode * Cam;
    SoDB::read(&in,Cam);

    if (!Cam){
        throw Base::Exception("Camera settings failed to read");
    }

    // toogle between persepective and orthographic camera
    if (Cam->getTypeId() != CamViewer->getTypeId())
    {
        _viewer->setCameraType(Cam->getTypeId());
        CamViewer = _viewer->getCamera();
    }

    SoPerspectiveCamera  * CamViewerP = 0;
    SoOrthographicCamera * CamViewerO = 0;

    if (CamViewer->getTypeId() == SoPerspectiveCamera::getClassTypeId()) {
        CamViewerP = (SoPerspectiveCamera *)CamViewer;  // safe downward cast, knows the type
    } else if (CamViewer->getTypeId() == SoOrthographicCamera::getClassTypeId()) {
        CamViewerO = (SoOrthographicCamera *)CamViewer;  // safe downward cast, knows the type
    }

    if (Cam->getTypeId() == SoPerspectiveCamera::getClassTypeId()) {
        if (CamViewerP){
            CamViewerP->position      = ((SoPerspectiveCamera *)Cam)->position;
            CamViewerP->orientation   = ((SoPerspectiveCamera *)Cam)->orientation;
            CamViewerP->nearDistance  = ((SoPerspectiveCamera *)Cam)->nearDistance;
            CamViewerP->farDistance   = ((SoPerspectiveCamera *)Cam)->farDistance;
            CamViewerP->focalDistance = ((SoPerspectiveCamera *)Cam)->focalDistance;
        } else {
            throw Base::Exception("Camera type mismatch");
        }
    } else if (Cam->getTypeId() == SoOrthographicCamera::getClassTypeId()) {
        if (CamViewerO){
            CamViewerO->viewportMapping  = ((SoOrthographicCamera *)Cam)->viewportMapping;
            CamViewerO->position         = ((SoOrthographicCamera *)Cam)->position;
            CamViewerO->orientation      = ((SoOrthographicCamera *)Cam)->orientation;
            CamViewerO->nearDistance     = ((SoOrthographicCamera *)Cam)->nearDistance;
            CamViewerO->farDistance      = ((SoOrthographicCamera *)Cam)->farDistance;
            CamViewerO->focalDistance    = ((SoOrthographicCamera *)Cam)->focalDistance;
            CamViewerO->aspectRatio      = ((SoOrthographicCamera *)Cam)->aspectRatio ;
            CamViewerO->height           = ((SoOrthographicCamera *)Cam)->height;
        } else {
            throw Base::Exception("Camera type mismatch");
        }
    }

    return true;
}

void View3DInventor::toggleClippingPlane()
{
    _viewer->toggleClippingPlane();
}

bool View3DInventor::hasClippingPlane() const
{
    return _viewer->hasClippingPlane();
}

void View3DInventor::setCursor(const QCursor& aCursor)
{
    _viewer->getWidget()->setCursor(aCursor);
}

void View3DInventor::dump(const char* filename)
{
    SoGetPrimitiveCountAction action;
    action.setCanApproximate(true);
    action.apply(_viewer->getSceneGraph());

    if ( action.getTriangleCount() > 100000 || action.getPointCount() > 30000 || action.getLineCount() > 10000 )
        _viewer->dumpToFile(filename,true);
    else
        _viewer->dumpToFile(filename,false);
}

void View3DInventor::windowStateChanged(MDIView* view)
{
    bool canStartTimer = false;
    if (this != view) {
        // If both views are child widgets of the workspace and view is maximized this view 
        // must be hidden, hence we can start the timer.
        // Note: If view is top-level or fullscreen it doesn't necessarily hide the other view
        // e.g. if it is on a second monitor.
        canStartTimer = (!this->isTopLevel() && !view->isTopLevel() && view->isMaximized());
    } else if (isMinimized()) {
        // I am the active view but minimized
        canStartTimer = true;
    }

    if (canStartTimer) {
        // do a single shot event (maybe insert a checkbox in viewer settings)
        int msecs = hGrp->GetInt("stopAnimatingIfDeactivated", 3000);
        if (!stopSpinTimer->isActive() && msecs >= 0) { // if < 0 do not stop rotation
            stopSpinTimer->setSingleShot(true);
            stopSpinTimer->start(msecs);
        }
    } else if (stopSpinTimer->isActive()) {
        // If this view may be visible again we can stop the timer
        stopSpinTimer->stop();
    }
}

void View3DInventor::stopAnimating()
{
    _viewer->stopAnimating();
}

/**
 * Drops the event \a e and writes the right Python command.
 */
void View3DInventor::dropEvent (QDropEvent * e)
{
    const QMimeData* data = e->mimeData();
    if (data->hasUrls()) {
        QList<QUrl> uri = data->urls();
        App::Document* pDoc = getAppDocument();
        if (pDoc) {
            for (QList<QUrl>::ConstIterator it = uri.begin(); it != uri.end(); ++it) {
                QFileInfo info((*it).toLocalFile());
                if ( info.exists() && info.isFile() ) {
                    if (info.isSymLink())
                        info.setFile(info.readLink());
                    // First check the complete extension
                    if (App::GetApplication().hasOpenType(info.completeSuffix().toAscii()))
                        Application::Instance->import(info.absoluteFilePath().toUtf8(), pDoc->getName());
                    // Don't get the complete extension
                    else if (App::GetApplication().hasOpenType( info.suffix().toAscii()))
                        Application::Instance->import(info.absoluteFilePath().toUtf8(), pDoc->getName());
                }
            }
        }
    } else {
        MDIView::dropEvent(e);
    }
}

void View3DInventor::dragEnterEvent (QDragEnterEvent * e)
{
    // Here we must allow uri drafs and check them in dropEvent
    const QMimeData* data = e->mimeData();
    if ( data->hasUrls() )
        e->accept();
    else
        e->ignore();
}

void View3DInventor::setCurrentViewMode(ViewMode newmode)
{
    ViewMode oldmode = MDIView::currentViewMode();
    MDIView::setCurrentViewMode(newmode);

    // This widget becomes the focus proxy of the embedded GL widget if we leave 
    // the 'Child' mode. If we reenter 'Child' mode the focus proxy is reset to 0.
    // If we change from 'TopLevel' mode to 'Fullscreen' mode or vice versa nothing
    // happens.
    // Grabbing keyboard when leaving 'Child' mode (as done in a recent version) should
    // be avoided because when two or more windows are either in 'TopLevel' or 'Fullscreen'
    // mode only the last window gets all key event even if it is not the active one.
    //
    // It is important to set the focus proxy to get all key events otherwise we would loose
    // control after redirecting the first key event to the GL widget.
    // We redirect these events in keyPressEvent() and keyReleaseEvent().
    if (oldmode == Child) {
        // To make a global shortcut working from this window we need to add
        // all existing actions from the mainwindow and its sub-widgets 
        QList<QAction*> acts = getMainWindow()->findChildren<QAction*>();
        this->addActions(acts);
        _viewer->getGLWidget()->setFocusProxy(this);
        // To be notfified for new actions
        qApp->installEventFilter(this);
    } else if (newmode == Child) {
        _viewer->getGLWidget()->setFocusProxy(0);
        qApp->removeEventFilter(this);
        QList<QAction*> acts = this->actions();
        for (QList<QAction*>::Iterator it = acts.begin(); it != acts.end(); ++it)
            this->removeAction(*it);
    }
}

bool View3DInventor::eventFilter(QObject* watched, QEvent* e)
{
    // As long as this widget is a top-level window (either in 'TopLevel' or 'FullScreen' mode) we
    // need to be notified when an action is added to a widget. This action must also be added to 
    // this window to allow to make use of its shortcut (if defined).
    // Note: We don't need to care about removing an action if its parent widget gets destroyed.
    // This does the action itself for us.
    if (watched != this && e->type() == QEvent::ActionAdded) {
        QActionEvent* a = static_cast<QActionEvent*>(e);
        QAction* action = a->action();

        if (!action->isSeparator()) {
            QList<QAction*> actions = this->actions();
            if (!actions.contains(action))
                this->addAction(action);
        }
    }

    return false;
}

void View3DInventor::keyPressEvent (QKeyEvent* e)
{
    ViewMode mode = MDIView::currentViewMode();
    if (mode != Child) {
        // If the widget is in fullscreen mode then we can return to normal mode either
        // by pressing the matching accelerator or ESC. 
        if (e->key() == Qt::Key_Escape) {
            setCurrentViewMode(Child);
        } else {
            // Note: The key events should be redirected directly to the GL widget and not to the main window
            // otherwise the first redirected key event always disappears in hyperspace.
            //
            // send the event to the GL widget that converts to and handles an SoEvent
            QWidget* w = _viewer->getGLWidget();
            QApplication::sendEvent(w,e);
        }
    } else {
        QMainWindow::keyPressEvent(e);
    }
}

void View3DInventor::keyReleaseEvent (QKeyEvent* e)
{
    ViewMode mode = MDIView::currentViewMode();
    if (mode != Child) {
        // send the event to the GL widget that converts to and handles an SoEvent
        QWidget* w = _viewer->getGLWidget();
        QApplication::sendEvent(w,e);
    } else {
        QMainWindow::keyReleaseEvent(e);
    }
}

void View3DInventor::focusInEvent (QFocusEvent * e)
{
    _viewer->getGLWidget()->setFocus();
}

#include "moc_View3DInventor.cpp"
