/***************************************************************************
 *   Copyright (c) 2005 Werner Mayer <werner.wm.mayer@gmx.de>              *
 *                                                                         *
 *   This file is part of the FreeCAD CAx development system.              *
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Library General Public           *
 *   License as published by the Free Software Foundation; either          *
 *   version 2 of the License, or (at your option) any later version.      *
 *                                                                         *
 *   This library  is distributed in the hope that it will be useful,      *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Library General Public License for more details.                  *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this library; see the file COPYING.LIB. If not,    *
 *   write to the Free Software Foundation, Inc., 59 Temple Place,         *
 *   Suite 330, Boston, MA  02111-1307, USA                                *
 *                                                                         *
 ***************************************************************************/


#include "PreCompiled.h"

#include <Base/Console.h>
#include <Base/Exception.h>
#include "WorkbenchManager.h"
#include "Workbench.h"
#include "MenuManager.h"
#include "ToolBarManager.h"
#include "CommandBarManager.h"
#include "DockWindowManager.h"
#include "MainWindow.h"

using namespace Gui;

WorkbenchManager* WorkbenchManager::_instance = 0;

WorkbenchManager* WorkbenchManager::instance()
{
    if (_instance == 0)
        _instance = new WorkbenchManager;
    return _instance;
}

void WorkbenchManager::destruct()
{
    delete _instance;
    _instance = 0;
}

WorkbenchManager::WorkbenchManager() : _activeWorkbench(0)
{
}

WorkbenchManager::~WorkbenchManager()
{
    for (QMap<QString, Workbench*>::Iterator it = _workbenches.begin(); it != _workbenches.end(); ++it) {
        Workbench* wb = it.value();
        delete wb;
    }

    MenuManager::destruct();
    ToolBarManager::destruct();
    CommandBarManager::destruct();
    DockWindowManager::destruct();
}

Workbench* WorkbenchManager::createWorkbench ( const QString& name, const QString& className )
{
    Workbench* wb = getWorkbench(name);

    if (!wb) {
        // try to create an instance now
        wb = (Workbench*) Base::Type::createInstanceByName((const char*)className.toAscii(),false);
        if (wb) {
            if (!wb->getTypeId().isDerivedFrom(Gui::Workbench::getClassTypeId())) {
                delete wb;
                QString error = QString("'%1' is not a workbench type").arg(className);
                throw Base::Exception((const char*)error.toAscii());
            }

            wb->setName( name );
            _workbenches[ name ] = wb;
        }
        else
            Base::Console().Log("WorkbenchManager::createWorkbench(): Can not create "
                "Workbench instance with type: %s\n",(const char*)className.toAscii());
    }

    return wb;
}

Workbench* WorkbenchManager::getWorkbench ( const QString& name )
{
    Workbench* wb=0;

    QMap<QString, Workbench*>::Iterator it = _workbenches.find(name);
    if (it != _workbenches.end()) {
        // returns the already created object
        wb = it.value();
    }

    return wb;
}

bool WorkbenchManager::activate( const QString& name, const QString& className )
{
    Workbench* wb = createWorkbench(name, className);
    if (wb) {
        _activeWorkbench = wb;
        wb->activate();
        return true;
    }
  
    return false;
}

Workbench* WorkbenchManager::active() const
{
    return _activeWorkbench;
}

QStringList WorkbenchManager::workbenches() const
{
    QStringList wb;
    for (QMap<QString, Workbench*>::ConstIterator it = _workbenches.begin(); it != _workbenches.end(); ++it)
        wb << it.key();
    return wb;
}
